/*  serial port interface for PS/2 Keyboard, Mouse and SFC Joypad 
 *  programed by piro
 *
 *  Based on ser_code_ksim v1.1.0
 *  serial port interface for Palm Portable Keyboard
 *  Copyright (C) 2003 
 *  Trap <CXL00145@nifty.com> (linux zaurus SL-C7x0 version)
 *
 *  Based on ser_code.c created by ...
 *    Dan Morris <dmorris@cs.brown.edu> (initial and windows version)
 *    Ralf Ackermann <rac@KOM.tu-darmstadt.de>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 *
 *  First release 2004.4.10 v0.1.0
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <unistd.h>
#include <libgen.h>
#include <fcntl.h>
#include <sys/ioctl.h>
#include <sys/wait.h>
#include <sys/types.h>
#include <termio.h>
#include <error.h>

// mouse
#include <linux/fb.h>
#include <asm/page.h>        
#include <sys/mman.h>
// end_mouse

#ifndef errno
extern int errno;
#endif

#include <asm/sharp_char.h>
#include <asm/sharp_keycode.h>
#include <linux/kbd_ll.h>


// Constants

#define MYFLG_SHIFT 0x01
#define MYFLG_CTRL 0x02
#define MYFLG_FN 0x04
#define MYFLG_CMD 0x08

#define RETOUR_OK 0
#define RETOUR_ERROR -1

#define RTS_TIMEOUT 600

#define VERSION "0.1.0"

#define KBDDRV "/dev/kbd"
#define VNCDRV "/dev/kbdsim"
#define SHARPKBDDRV "/dev/sharp_kbdctl"
#define TSDRV "/dev/ts"
#define KMAP_ORG    "loadkmap < /etc/slc.kmap"
#define KMAP_PS2    "loadkmap < /etc/ps2.kmap"

// user structure

typedef struct key_mapping {
  char *keyname;
  long keycode;
} key_mappings;

// global variables

static int fdd = -1;                // IOCTL ԤǥХѤΥեϥɥ
static int fds = -1;                // ܡɤμѥեϥɥ
static int infoflag = 0;                // debug åɽե饰
static int vncflag = 0;                // VNCɥ饤ѥ⡼ɤΥե饰
static int keyrepeat = 0;                // ԡȤΤΥ
static int keyrep_first = 3;        // 󥭡ԡԤ(0.3s)
static int keyrep_next = 1;        // 2ܰʹߥԡԤ(0.1s)
static long lastkeycode = -1;        // Ǹ˲줿Ф
static char * program_name = "";        // Υץ̾

static char * serialdevice = "/dev/ttyS1";
static char * keymapfile   = NULL;
static char * keymapfile_n   = "/opt/QtPalmtop/etc/ser_code_kmjsim.ini";
static char * keymapfile_c   = "/etc/ser_code_kmjsim_console.ini";

static char * metas_str[8] = { "lshift", "lctrl", "lalt", "rcmd", "rshift", "rctrl", "ralt", "rcmd" };

// Ǽΰ
static int metaflags;        // meta-key β֤򵭲ѿ
static char keys[128];        // ƥ ON/OFF ֤ݻ
static unsigned char    metas[256];        // Υɤ meta-key Ǥ뤫򼨤
static key_mappings    keymap[16][256][2];    // cmd, fn, ctrl, shift ȹ礻16̤ 256 ĤΥ & extend code

static int keymap_j[16];       //Joystick key map

// mouse
static int mode = 0;
static int consolemode = 0;
static int rotatemode = 0;

static FILE *fp_ser;
static FILE *fp_kbd;
static FILE *fp_ts;


static int console_fd;
static struct fb_var_screeninfo vinfo;
static struct fb_fix_screeninfo finfo;
static int mapped_memlen;
static int mapped_offset;
static short *mapped_mem;
static short *mapped_mem_start;
static short *mapped_mem_end;

int fb_offset;
//end_mouse


// prototypes

int read_key_mapping( char *map_file );
int read_key_mapping_direct( char *map_file );

void infoarea_init( void );
int check_meta( const char *str );
int read_key_mapping( char *map_file );

void rts_down_up( int fd );
int init_ser( char *ser_device ,int speed);
int set_ser_param( int fd ,int speed);
void flush_ser( int fd );

unsigned char get_1_char( void );

void sig_handler( int status );

void close_all( void );
void usage( int status );
void printversion( void );
void printinfo( void );

void onekey_v( int c, int pressflag );
void onekey_k( int c, int pressflag );
void onekey(int c,int pressflag);


//  mouse

#define POINTERCAL "/etc/pointercal"

static struct qt_cal {
  int ok;
  int a, b, c, d, e, f, s;
} cal;

static struct q_cal {
  int ok;
  double a, b, c, d, e, f, s;
} qcal;

static short pointer[16][16] = { 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0001, 0x0001, 0x0001, 0x0001, 0x0001, 0x0001, 0x0001,
                                 0x0000, 0xFFFF, 0xFFFF, 0xFFFF, 0xFFFF, 0xFFFF, 0xFFFF, 0xFFFF, 0x0000, 0x0001, 0x0001, 0x0001, 0x0001, 0x0001, 0x0001, 0x0001,
                                 0x0000, 0xFFFF, 0xFFFF, 0xFFFF, 0xFFFF, 0xFFFF, 0xFFFF, 0xFFFF, 0x0000, 0x0001, 0x0001, 0x0001, 0x0001, 0x0001, 0x0001, 0x0001,
                                 0x0000, 0xFFFF, 0xFFFF, 0xFFFF, 0xFFFF, 0xFFFF, 0xFFFF, 0xFFFF, 0x0000, 0x0001, 0x0001, 0x0001, 0x0001, 0x0001, 0x0001, 0x0001,
                                 0x0000, 0xFFFF, 0xFFFF, 0xFFFF, 0xFFFF, 0xFFFF, 0xFFFF, 0xFFFF, 0x0000, 0x0001, 0x0001, 0x0001, 0x0001, 0x0001, 0x0001, 0x0001,
                                 0x0000, 0xFFFF, 0xFFFF, 0xFFFF, 0xFFFF, 0xFFFF, 0xFFFF, 0xFFFF, 0x0000, 0x0001, 0x0001, 0x0001, 0x0001, 0x0001, 0x0001, 0x0001,
                                 0x0000, 0xFFFF, 0xFFFF, 0xFFFF, 0xFFFF, 0xFFFF, 0xFFFF, 0xFFFF, 0x0000, 0x0001, 0x0001, 0x0001, 0x0001, 0x0001, 0x0001, 0x0001,
                                 0x0000, 0xFFFF, 0xFFFF, 0xFFFF, 0xFFFF, 0xFFFF, 0xFFFF, 0xFFFF, 0xFFFF, 0x0000, 0x0001, 0x0001, 0x0001, 0x0001, 0x0001, 0x0001,
                                 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0xFFFF, 0xFFFF, 0xFFFF, 0xFFFF, 0xFFFF, 0x0000, 0x0001, 0x0001, 0x0001, 0x0001, 0x0001,
                                 0x0001, 0x0001, 0x0001, 0x0001, 0x0001, 0x0000, 0xFFFF, 0xFFFF, 0xFFFF, 0xFFFF, 0xFFFF, 0x0000, 0x0001, 0x0001, 0x0001, 0x0001,
                                 0x0001, 0x0001, 0x0001, 0x0001, 0x0001, 0x0001, 0x0000, 0xFFFF, 0xFFFF, 0xFFFF, 0xFFFF, 0xFFFF, 0x0000, 0x0001, 0x0001, 0x0001,
                                 0x0001, 0x0001, 0x0001, 0x0001, 0x0001, 0x0001, 0x0001, 0x0000, 0xFFFF, 0xFFFF, 0xFFFF, 0xFFFF, 0xFFFF, 0x0000, 0x0001, 0x0001,
                                 0x0001, 0x0001, 0x0001, 0x0001, 0x0001, 0x0001, 0x0001, 0x0001, 0x0000, 0xFFFF, 0xFFFF, 0xFFFF, 0xFFFF, 0xFFFF, 0x0000, 0x0001,
                                 0x0001, 0x0001, 0x0001, 0x0001, 0x0001, 0x0001, 0x0001, 0x0001, 0x0001, 0x0000, 0xFFFF, 0xFFFF, 0xFFFF, 0x0000, 0x0001, 0x0001,
                                 0x0001, 0x0001, 0x0001, 0x0001, 0x0001, 0x0001, 0x0001, 0x0001, 0x0001, 0x0001, 0x0000, 0xFFFF, 0x0000, 0x0001, 0x0001, 0x0001,
                                 0x0001, 0x0001, 0x0001, 0x0001, 0x0001, 0x0001, 0x0001, 0x0001, 0x0001, 0x0001, 0x0001, 0x0000, 0x0001, 0x0001, 0x0001, 0x0001 };

static short prev_image1[16][16];
static short prev_image2[16][16];
static short prev_x,prev_y;
static int st=0;

void mouseon(int x,int y)
{
  int i,j;
  
  x -= 16;
  for (i=0 ; i<16; i++){
    for (j=0 ; j<16; j++){
      if((mapped_mem+i+x+(j+y)*finfo.line_length/2) > mapped_mem_start &&
         (mapped_mem+i+x+(j+y)*finfo.line_length/2) < mapped_mem_end &&
         ( (i + x) < 480 ) && ( (i + x) >= 0 ) && ( (j + y) < 640 ) && ( (j + y) >= 0 )  ){
        if (pointer[15-i][j] != 1) {
          *(mapped_mem+i+x+(j+y)*finfo.line_length/2 ) = pointer[15-i][j];
        }
      }
    }
  }
  
  return;
}

void mouseoff(int x,int y)
{
  int i,j;
  
  x -= 16;
  for (i=0 ; i<16; i++){
    for (j=0 ; j<16; j++){
      if((mapped_mem+i+x+(j+y)*finfo.line_length/2) > mapped_mem_start &&
         (mapped_mem+i+x+(j+y)*finfo.line_length/2) < mapped_mem_end && 
         ( (i + x) < 480 ) && ( (i + x) >= 0 ) && ( (j + y) < 640 ) && ( (j + y) >= 0 ) &&
     *(mapped_mem+i+x+(j+y)*finfo.line_length/2 ) == prev_image2[i][j] ) {  
    *(mapped_mem+i+x+(j+y)*finfo.line_length/2 ) = prev_image1[i][j];
      }
    }
  }
}

void grabrect1(int x,int y)
{
  int i,j;
  
  x -= 16;
  for (i=0 ; i<16; i++){
    for (j=0 ; j<16; j++){
      if((mapped_mem+i+x+(j+y)*finfo.line_length/2) > mapped_mem_start &&
         (mapped_mem+i+x+(j+y)*finfo.line_length/2) < mapped_mem_end &&
         ( (i + x) < 480 ) && ( (i + x) >= 0 ) && ( (j + y) < 640 ) && ( (j + y) >= 0 )  ){
        prev_image1[i][j]=*(mapped_mem+i+x+(j+y)*finfo.line_length/2 );
      }
    }
  }
}

void grabrect2(int x,int y)
{
  int i,j;
  
  x -= 16;
  for (i=0 ; i<16; i++){
    for (j=0 ; j<16; j++){
      if((mapped_mem+i+x+(j+y)*finfo.line_length/2) > mapped_mem_start &&
         (mapped_mem+i+x+(j+y)*finfo.line_length/2) < mapped_mem_end &&
         ( (i + x) < 480 ) && ( (i + x) >= 0 ) && ( (j + y) < 640 ) && ( (j + y) >= 0 )  ){
        prev_image2[i][j]=*(mapped_mem+i+x+(j+y)*finfo.line_length/2 );
      }
    }
  }
}

void loadConfig()
{
  FILE *pcal;
  cal.ok = 0;
  pcal = fopen(POINTERCAL, "r");
  if(pcal == NULL){
    fprintf(stderr, "Warning: cannot open " POINTERCAL ".\n");
  }else{
    int n;
    n = fscanf(pcal, "%d %d %d %d %d %d %d",
               &cal.a, &cal.b, &cal.c, &cal.d, &cal.e, &cal.f, &cal.s);
    qcal.a = cal.a /1000;
    qcal.b = cal.b /1000;
    qcal.c = cal.c /1000;
    qcal.d = cal.d /1000;
    qcal.e = cal.e /1000;
    qcal.f = cal.f /1000;
    qcal.s = cal.s /1000;
    if(n != 7){
      fprintf(stderr, "Warning: " POINTERCAL " is unknown format.\n");
    }else{
      cal.ok = 1;
    }
    fclose(pcal);
  }
  
  
  console_fd = open("/dev/fb0", O_RDWR, 0);
  if ( console_fd < 0 ) {
    return;
  }
  
  if ( ioctl(console_fd, FBIOGET_FSCREENINFO, &finfo) < 0 ) {
    printf("error");
    return;
  }
  
  if ( ioctl(console_fd, FBIOGET_VSCREENINFO, &vinfo) < 0 ) {
    printf("error");
    return;
  }
  
  mapped_offset = (((long)finfo.smem_start) -
                   (((long)finfo.smem_start)&~(PAGE_SIZE-1)));
  mapped_memlen = finfo.smem_len+mapped_offset;
  mapped_mem = (short*)mmap(NULL, mapped_memlen,PROT_READ|PROT_WRITE, MAP_SHARED, console_fd, 0);
  mapped_mem_start = mapped_mem;
  mapped_mem_end = mapped_mem + finfo.smem_len;
  
  close(console_fd);
}

// end mouse

/*
  Ǽΰ̥륯ꥢ
*/
void infoarea_init( void )
{
  metaflags = 0;
  memset( keys, 0, sizeof(keys) );
  memset( metas, 0, sizeof(metas) );
  memset( keymap, 0, sizeof(keymap) );
}

/*
  Ϳ줿ʸ󤬡ɤ meta-key ˰פ뤫Ĵ١
  metaflags ͤ(bitñ̤Υե饰)
  
  ret    == 0 : meta-key ǤϤʤ
  != 0 : meta-key Ǥ롣ͤ bitñ̤Υե饰͡
*/
int check_meta( const char *str )
{
  int i;
  int flags = 0;
  
  for( i = 0; i < sizeof(metas_str)/sizeof(metas_str[0]); i++ ) {
    if ( strcasecmp( str, metas_str[i] ) == 0 ) {
      flags = 1 << i;
    }
  }
  
  return( flags );
}



/*
  ե뤫顢ޥåԥ󥰤ɤ߹
  
  ret    RETOUR_OK    ｪλ(ޥåץեȤϤȤ⤫)
  RETOUR_ERROR    ۾ｪλ(ޥåץե뤬ʤ)
*/
int read_key_mapping( char *map_file )
{
  int flags;
  int rcnt;
  unsigned int id;
  long keycode;
  FILE *in_file;
  char line[80];
  char keyname[80];
  char flagstr[80];
  int extend;
  
  // ѿꥢ
  infoarea_init();
  memset( line, 0, sizeof(line) );
  
  // ե򳫤
  if ( NULL == (in_file = fopen( map_file, "r" )) ) {
    return( RETOUR_ERROR );
  }
  
  // ե1Ԥɤߤʤ
  while ( ! feof(in_file) ) {
    
    // 1ɤ߹ (line byteʾԵ)
    if ( NULL != fgets(line, sizeof(line)-1, in_file) ) {
      
      // Ƭ # ιԤϥȹԤȤ̵뤹롣
      if ( line[0] == '#' ) continue;
      
      // meta-key ԤΤĤɤ߹Ǥߤ
      rcnt = sscanf(line, "%s %lx", keyname, &id );
      
      // Ƭιܤʤ̵뤷ƼιԤ(Ԥʤ)
      if ( rcnt < 1 ) continue;
      
      // ιԤ meta-key ԤǤСɤ߼ä metas ꤹ롣
      if ( (flags = check_meta( keyname )) != 0 ) {
	if ( (rcnt == 2) && (id < 256) ) metas[ id ] = flags;
	continue;
      }
      
      // ¾Υѥ᡼
      if ( strcasecmp( keyname, "keyrep_first" ) == 0 ) {
	keyrep_first = id & 0x7f;
	continue;
      }
      
      if ( strcasecmp( keyname, "keyrep_next" ) == 0 ) {
	keyrep_next = id & 0x7f;
	continue;
      }
      
      if ( strcmp(keyname,"J") == 0){
        rcnt = sscanf( line, "%*s %d %s %lx", &id, keyname, &keycode );
        if ( rcnt < 3 ) continue;
        if ( id > 256 ) continue;
        keymap_j[id] = keycode;
        continue;
      }      
      
      // Ƭܤʸ 16ʿʸǤʤȴ롣
      if (  (strlen(keyname) != 2) || (! isxdigit(keyname[0])) || (! isxdigit(keyname[1])) ){
        if ((strlen(keyname) != 4) || (! isxdigit(keyname[0])) || (! isxdigit(keyname[1]))
	    || (! isxdigit(keyname[2])) || (! isxdigit(keyname[3])) ) {
          continue; 
        }
      }
      
      if ( strlen(keyname) == 4) {
        if ( (( keyname[0] != 'E') && ( keyname[0] != 'e') ) || ( keyname[1] != '0') ){
	  continue; 
	}
        extend = 1;
      }else{
        extend = 0;
      }
      
      
      
      // ԤȤƲեޥåȤɤ߼롢
      rcnt = sscanf( line, "%x %s %s %lx", &id, keyname, flagstr, &keycode );
      
      id &= 0x00ff;
      
      // 4ɤʤФ
      if ( rcnt < 4 ) continue;
      
      // ϰϥå
      if ( id > 256 ) continue;
      
      // ե饰 S, C, F, M, N Σʸȹ礻롣 ex. SCF
      flags = 0;
      if ( strchr( flagstr, 'S' ) != NULL ) flags |= MYFLG_SHIFT;
      if ( strchr( flagstr, 'C' ) != NULL ) flags |= MYFLG_CTRL;
      if ( strchr( flagstr, 'F' ) != NULL ) flags |= MYFLG_FN;
      if ( strchr( flagstr, 'M' ) != NULL ) flags |= MYFLG_CMD;
      
      keymap[flags][id][extend].keyname    = strdup(keyname);
      keymap[flags][id][extend].keycode    = keycode;
    }
  }
  
  fclose( in_file );
  
  return( RETOUR_OK );
}

/*
  ꤵ줿ꥢݡȤ¾ץ
  
  in        char        *ser_device        ꥢǥХ
  
  ret        RETOUR_OK        
  RETOUR_ERROR        ץ or ¾å or ѥ᡼꼺
  
  Хѿ fdd  -1 ʳǤХץ󤵤Ƥ롣
   directflag 򻲾Ȥƥ⡼ɤѤƤ롣
*/
int init_ser( char *ser_device, int speed )
{
  // ser_device Υץ
  if ( (fds = open(ser_device, O_RDWR | O_NOCTTY, 0)) == -1 ) {
    printf( "Error opening serial device !\n" );
    return( RETOUR_ERROR );
  }
  
  // ¾ˤĤޤʤ褦¾å
  if ( flock( fds, LOCK_EX | LOCK_NB ) == -1 ) {
    return( RETOUR_ERROR );
  }
  
  // ѥ᡼
  if ( set_ser_param(fds,speed) == RETOUR_ERROR ) {
    printf( "Couldn't set serial parameters !\n" );
    return( RETOUR_ERROR );
  }else{
    return( RETOUR_OK );
  }
}


/*
  ꥢݡȤΥԡɤ¾
*/
int set_ser_param( int fd ,int speed)
{
  int line;
  struct termio sgtty;
  int speed_param;
  
  if (ioctl(fd, TCGETA, &sgtty) < 0) {
    printf( "can't TCGETA\n" );
    return( RETOUR_ERROR );
  }
  switch (speed) {
  case 19200:
    speed_param = B19200;
    break;
  case 38400:
    speed_param = B38400;
    break;
  case 57600:
    speed_param = B57600;
    break;
  default:
    speed_param = B9600;
    break;
  }
  
  
  sgtty.c_iflag =  (IMAXBEL | IGNBRK);
  sgtty.c_oflag =  0;
  sgtty.c_lflag =  0;
  sgtty.c_cflag =  (CLOCAL | HUPCL | CS8 | CREAD | speed_param);
  sgtty.c_cflag &= ~(PARENB | PARODD);
  sgtty.c_cc[VTIME] =  1;                /* ॢȤ 0.1 s        */
  sgtty.c_cc[VMIN] =  0;                /* ʸޤǥ֥å */
  
  if ( ioctl(fd, TCSETA, &sgtty) < 0 ) {
    printf( "can't TCSETA\n" );
    return( RETOUR_ERROR );
  }
  
  ioctl( fd, TIOCMGET, &line );
  line |= TIOCM_DTR;
  ioctl( fd, TIOCMSET, &line );
  
  if ( infoflag ) printf( "DTR set to HIGH\n" );
  
  sleep( 1 );
  flush_ser( fd );
  
  return( RETOUR_OK );
}


/*
  ϥɥΥեå
*/
void flush_ser( int fd )
{
  ioctl( fd, TCFLSH, 0 );
}


/*
  1ʸϴؿ
  
  ǥȥԡȽԤäƤ롣
*/
unsigned char get_1_char( void )
{
  unsigned char recv_char;
  int r_code;
  static int timeout = RTS_TIMEOUT;
  
  for(;;) {
    /* wait for a char to arrive */
    r_code = read( fds, &recv_char, 1 );
    
    if ( r_code > 0 ) break;                // 줿/Υ줿ϥ롼פȴ
    
  }
  
  return( recv_char );
}


// Control-C ʥϥɥ
void sig_handler( int status )
{
  printf( "SIGNAL(%d) ΤǥݡȤĤƽλޤ\n", status );
  exit( 1 );
}


/*
  ץλθ
*/
void close_all( void )
{
  int i;
  long keycode;
  
  // open ѤΥϥɥĤƽλ
  
  if(consolemode){
    system(KMAP_ORG);
  }
  // ꥢݡ
  if ( fds != -1 ) {
    close( fds );
    fds = -1;
  }
  
  if ( fp_ser != NULL ) {
    fclose( fp_ser );
  }

  if ( fp_kbd != NULL ) {
    fclose( fp_kbd );
  }

  if ( fp_ts != NULL ) {
    fclose( fp_ts );
  }
  
  if (consolemode != 1){
    if(st == 1){
      mouseoff(prev_y,prev_x);
    }
  }
  
}


/*
  ˡɽ
*/
void usage( int status )
{
  if ( status != 0 ) {
    printf( "-h for more information.\n" );
  }else{
    printf( "Usage: %s [OPTION]\n", program_name );
    
    printf(
           "-p : serial port device name (defalt: /dev/ttyS1)\n"
           "-s : serial port speed (defalt: 9600 bps)\n"
           "-V : vnc kbdsim module use\n"
           "-i : debug info print\n"
           "-b : background daemon\n"
           "-v : print version\n"
           "-h : display this message\n"
           "-m : key mapping file name\n"
           "-M : Use mouse\n"
           "-c : console/X mode\n"
           "     default: /opt/QtPalmtop/etc/ser_code_kmsim.ini \n"
           );
  }
  
  exit( status );
}


/*
  ץࡦСɽ
*/
void printversion( void )
{
  printf( "%s version %s\n", program_name, VERSION );
  exit( 0 );
}


/**************************************************************************************************************
        ե˴ŤơϽԤᥤ
***************************************************************************************************************/
int main( int argc, char **argv )
{
  unsigned char rawcode;
  
  int daemonflag = 0;
  int pressflag;
  int meta;
  int scancode;
  long keycode;
  char *keyname;
  char checksum;
  int leds = 0;
  
  int on_key;
  char key_code[10];
  int key_count = 0;
  int chr_count = 0;
  
  int use_mouse = 0;
  int mousecode = 0;
  int mouse_x = 0, mouse_y = 0, mouse_x_dif = 0, mouse_y_dif = 0;
  int mouse_x_sgn = 0, mouse_y_sgn = 0, mouse_x_of = 0, mouse_y_of = 0, mouse_mb = 0;
  int mouse_lb = 0, mouse_rb = 0, raw_x, raw_y; 
  
  int speed;
  
  // λˡץѥϥɥĤ뤿
  if ( atexit( close_all ) != 0 ) {
    usage( 2 );
  }
  
  // ư줿ʬȤ̾¸
  program_name = basename( argv[0] );
  
  
  // ϤơƼѿꤹ
  {
    int i;
    int j;
    int k;
    
    for( i = 1; i < argc; i++ ) {
      if ( argv[i][0] != '-' ) {
	usage( 1 );
      }else{
	k = i;
	for( j = 1; j < strlen(argv[k]); j++ ) {
	  switch( argv[k][j] ) {
	  case 'p':
	    if ( ++i < argc ) serialdevice = argv[i]; else usage(1);
	    break;
	  case 's':
	    if ( ++i < argc ) speed = atol(argv[i]); else usage(1);
	    break;
	  case 'm':
	    if ( ++i < argc ) keymapfile = argv[i]; else usage(1);
	    break;
	  case 'V': vncflag = 1; break;
	  case 'i': infoflag = 1; break;
	  case 'b': daemonflag = 1; break;
	  case 'M': use_mouse = 1; break;
	  case 'c': consolemode = 1; break;
	  case 'x': rotatemode = 1; break;
	  case 'v': printversion();
	  case 'h': usage( 0 );
	  default:  usage( 1 );
	  }
	}
      }
    }
  }
  
  // եɤ߹ࡣ
  if(consolemode){
    if ( read_key_mapping( keymapfile_c ) !=  RETOUR_OK ) {
      printf( "Key map file is not found\n" );
      exit( 1 );
    }
    system(KMAP_PS2);
  }else {
    if ( read_key_mapping( keymapfile_n ) !=  RETOUR_OK ) {
      printf( "Key map file is not found\n" );
      exit( 1 );
    }
  }

  // kill ˥ץѥݡȤĤ뤿᥷ʥϥɥ
  signal( SIGINT,  &sig_handler );
  signal( SIGTERM, &sig_handler );
  signal( SIGHUP,  &sig_handler );
  
  //mouse
  loadConfig();
  
  // ꥢݡȤΥץ
  if ( init_ser( serialdevice, speed ) != RETOUR_OK ) {
    exit( 1 );
  }
  if( (fp_ser = fopen(serialdevice, "w")) == NULL)
      exit(1);

  if (use_mouse){
    fprintf(fp_ser, "e");
    fflush(fp_ser);
    if( (fp_ts = fopen(TSDRV, "w")) == NULL)
      exit(1);
  } 

  if (vncflag){  
    if( (fp_kbd = fopen(VNCDRV, "w")) == NULL)
      exit(1);
  }else {
    if( (fp_kbd = fopen(KBDDRV, "w")) == NULL)
      exit(1);
  }
  
  // daemon
  if ( daemonflag ) {
    infoflag = 0;
    daemon( 0, 0 );
  }
  
  for(;;) {
    
    // 1ʸ Ͻ///////////////////////////////////////////////////////////////////////////////////
    rawcode = get_1_char();
    
//    printf("%02x \n",rawcode);
//    fflush(stdout);
    if (chr_count == 0) {
      switch (rawcode){
      case 'K':
	mode = 0;
        break;
      case 'M':
        if(use_mouse)
	  mode = 1;
        break;
      case 'J':
	mode = 2;
        break;
      default:
        continue;
        break;
      }
    }
    
    if (chr_count < 5) {
      key_code[chr_count] = rawcode;
      chr_count++;
      continue;
    }
    key_code[chr_count] = rawcode;
    
    chr_count = 0;

    if (((key_code[1]+key_code[2]+key_code[3]+key_code[4]) & 0xff) != key_code[5] ) {
      printf("Error checksum\n");
      continue;
    }
    
    if (mode == 0 ){
      // 󥳡Τ߼Ф
      int extend = 0;
      static int extend_2nd = 0;
      
      if(extend_2nd !=0){
	extend_2nd--;
	continue;
      }
      scancode = key_code[key_code[1]+1];

      // brake code Υ줿, ʤХ줿
      keys[ scancode ] = pressflag = (key_code[key_code[1]] == 0xF0) ? 0 : 1;
      
      if (key_code[2] == 0xe1) {
	extend_2nd = 3;
	extend = 1;
	scancode = 0x80;
      }
      if (key_code[2] == 0xe0) {
	extend = 1;
      }
      
      
      // Shift, Ctrl, Fn, Cmd ʤɤ metaˤĤե饰ꤹ롣
      meta = metas[ scancode ];
      
      if ( meta != 0 ) {
	if ( pressflag ) {
	  metaflags |= meta;
	}else{
	  metaflags &= ~meta;
	}
	chr_count = 0;
      }
      
      // 겼ϡ̾省ν
      
      // ߤ Shift, Ctrl, Fn, Cmd б ͤ
      meta = (metaflags & 0x0f) | (metaflags >> 4);
      
      keycode = keymap[meta][scancode][extend].keycode;
      keyname = keymap[meta][scancode][extend].keyname;
      
      //      printf("scancode %08x keycode %08x extend %d meta %08x keyname %s\n",scancode,keycode,extend,meta,keyname);
      
      // keycode  0 ξϡб륭̵Τǲ⤷ʤ
      if (keyname == NULL)
	continue;
      
      if ((strcmp(keyname,"NumLock") == 0) && pressflag){
	if((leds & 0x02 )== 0)
          leds |= 2;
	else
	  leds &= ~2;
	
	fprintf(fp_ser,"k\xed%c",leds);
	fflush(fp_ser);
      }
      
      if ((strcmp(keyname,"Caps") == 0) && pressflag){
	if((leds & 0x04 )== 0)
	  leds |= 4;
	else
	  leds &= ~4;
	fprintf(fp_ser,"k\xed%c",leds);
	fflush(fp_ser);
      }
      
      if ( keycode == 0 ) {
	chr_count = 0;
	continue;
      }
      
      if((keycode & 0xff00) == 0x1e00) {
	if(pressflag) {
	  if(meta & 0x01){
 	    onekey(27,0);
	  }
	  onekey(30,1);
	}
	onekey((keycode & 0xff),pressflag);
	
	if(!pressflag){
	  if(meta & 0x01){
 	    onekey(27,1);
	  }
	  onekey(30,0);
	}
      }else{    
	onekey((keycode & 0xff), pressflag);
      }
      //      printf(" %03d %03d \n",(keycode & 0xff), pressflag);
      
    }else if(mode == 1){
      // Mouse Control///////////////////////////////////////////////////////////////////////////////

      mouse_y_of = (key_code[2] >> 7) & 0x01;
      mouse_x_of = (key_code[2] >> 6) & 0x01;
      mouse_y_sgn = (key_code[2] >> 5) & 0x01;
      mouse_x_sgn = (key_code[2] >> 4) & 0x01;
      mouse_mb = (key_code[2] >> 2) & 0x01;
      mouse_rb = (key_code[2] >> 1) & 0x01;
      mouse_lb = (key_code[2] >> 0) & 0x01;
      mouse_x_dif = key_code[3];
      mouse_y_dif = key_code[4];
      
      mouse_x = mouse_x + mouse_x_dif;
      mouse_y = mouse_y + mouse_y_dif;
      if (mouse_x_sgn)
	mouse_x -= 256;
      if (mouse_y_sgn)
	mouse_y -= 256;
      if(mouse_x > 639)
	mouse_x = 640;
      if(mouse_y > 479)
	mouse_y = 479;
      if(mouse_x < 0)
	mouse_x = 0;
      if(mouse_y < 0)
	mouse_y = 0;
      
      //        printf(" %03d %03d %03d ",mouse_x,mouse_y,mouse_lb);
      //        fflush(stdout);
      raw_y = (mouse_y * qcal.s * qcal.e - qcal.c * qcal.e) /(qcal.e * qcal.a);
      raw_x = (mouse_x * qcal.s * qcal.a - qcal.a * qcal.f) /(qcal.e * qcal.a);
      if(rotatemode == 0){
	fprintf(fp_ts, "%d %d %d 0\n",mouse_lb*500,raw_y,raw_x);
	fflush(fp_ts);
	//          printf("%03d %03d %03d 0\r",mouse_lb*500,raw_y,raw_x);
      }else{
	fprintf(fp_ts, "%d %d %d 0\n",(mouse_lb+1)*500,raw_x,480-raw_y);
	fflush(fp_ts);
	//          printf("%03d %03d %03d 0\r",(mouse_lb+1)*500,raw_x,480-raw_y);
      }
      if(consolemode == 0){
	if(st == 1){
	  mouseoff(prev_y,prev_x);
	}
	grabrect1(mouse_y,mouse_x);
	st = 1;
	mouseon(mouse_y,mouse_x);
	grabrect2(mouse_y,mouse_x);
	prev_x =mouse_x;
	prev_y =mouse_y;
	
      }
    }else if(mode == 2){
      // joystick///////////////////////////////////////////////////////////////////////////////////////
      static int pre_code = 0;
      int code;
      int i;
      
      code = (key_code[2] << 8) + key_code[3];
      
      //      printf("Joystick %04x\n",code);
      //printf("Joystick %04x\n",code ^ pre_code);
      
      for (i = 0; i<16; i++){
        if((code ^ pre_code) & ( 1 << i ) ){
	  onekey((keymap_j[i] & 0xff), (code >> i & 0x0001));
	  //          printf("code %08x press %d\n",keymap_j[i], (code >> i & 0x0001) );
        }
      }
      pre_code = code;
      
    }    
  }
  exit( 0 );                // ɬפʥݡȤΥϡatexit 롼ǹԤ
}

/*
  VNCɥ饤ѤΥϽ
*/
void onekey_v( int c, int pressflag )
{
  unsigned char buf[3];
  
  buf[0] = c; buf[1] = pressflag; buf[2] = '\0';
  fprintf(fp_kbd, "%c%c",c , pressflag);
}

/*
  Special Kernel /dev/kbd ( mknod /dev/kbd c 241 0 )
*/
void onekey_k( int c, int pressflag )
{
  fprintf(fp_kbd, " %d %d \n",c , pressflag);
  fflush(fp_kbd);
  
}
/*
  Ͻ()
*/
void onekey( int c, int pressflag )
{
  if(vncflag)
    onekey_v( c, pressflag );
  else
    onekey_k( c, pressflag);
}
